{% extends "!classic/layout.html" %}
{%- block doctype -%}
<!DOCTYPE html>
{%- endblock %}

{% block extrahead %}
    <meta http-equiv="X-UA-Compatible" content="IE=Edge" />
    <link rel="shortcut icon" type="image/png" href="{{ pathto('_static/mini_logo.png', 1) }}" />
    <meta name="viewport" content="width=device-width,initial-scale=0.8">
    {% if not theme_embedded %}
      {{ super() }}
      <script type="text/javascript" src="{{ pathto('_static/copybutton.js', 1) }}"></script>
      <script type="text/javascript" src="{{ pathto('_static/sidebar.js', 1) }}"></script>
    {% endif %}
    {% if theme_googlewebfonturl %}
      <link rel="stylesheet" type="text/css"
        href="{{ theme_googlewebfonturl }}">

    {% endif %}
{% endblock %}

{%- block sidebarlogo %}
    <a id="sidebar-anchor"></a>
    {{ super() }}
{% endblock %}

{%- block relbar1 %}
    <div class="related">
      <h3>{{ _('Navigation') }}</h3>
      <ul>
        {%- block rootrellink %}
        <li class="responsive-menu"><a href="#sidebar-anchor" title="Navigation">&#9776;</a></li>
        <li><a href="{{ pathto(master_doc) }}">{{ shorttitle|e }}</a>{{ reldelim1 }}</li>
        {%- endblock %}
        {%- for parent in parents %}
          <li><a href="{{ parent.link|e }}" {% if loop.last %}{{ accesskey("U") }}{% endif %}>{{ parent.title }}</a>{{ reldelim1 }}</li>
        {%- endfor %}
        {%- block relbaritems %} {% endblock %}
      </ul>
    </div>
    {% if theme_githuburl %}
        <div class="badge">
            <a href="{{ theme_githuburl }}">Fork me on GitHub</a>
            <img src="{{ pathto('_static/right-red@2x.png', 1) }}">
        </div>
    {% endif %}
{% endblock %}

{%- block relbar2 %}
    <div class="related">
      <h3>{{ _('Navigation') }}</h3>
      <ul>
        {%- for rellink in rellinks %}
        <li class="right" {% if loop.first %}style="margin-right: 10px"{% endif %}>
          <a href="{{ pathto(rellink[0]) }}" title="{{ rellink[1]|striptags|e }}"
             {{ accesskey(rellink[2]) }}>{{ rellink[3] }}</a>
          {%- if not loop.first %}{{ reldelim2 }}{% endif %}</li>
        {%- endfor %}
      </ul>
    </div>
{% endblock %}
